/**
 * Copyright 2012, Ayuntamiento de Tarragona
 *  
 * Desarrollador por  Guadaltel, S.A.
 * 
 * El presente cdigo fue desarrollado con financiacin del Ministerio 
 * de Industria Turismo y Comercio con el nmero de proyecto TSI-050100-2009-219.
 * 
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by the 
 * European Commission  subseqent versions of the EUPL (the "Licence");  you 
 * may not use this work except in compliance with the Licence.
 *
 * You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
 *
 * Unless required by applicable law or agreed to in writing, software distributed 
 * under the  Licence is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR 
 * CONDITIONS OF ANY KIND, either express or implied. See the Licence for the 
 * specific language governing  permissions and limitations under the Licence.
 */
package es.guadaltel.framework.authenticator.impl;

import es.guadaltel.framework.authenticator.Authenticator;
import es.guadaltel.framework.authenticator.ICertReader;
import es.guadaltel.framework.authenticator.domain.InfoCertificado;
import es.guadaltel.framework.authenticator.exception.AuthenticatorException;
import es.guadaltel.framework.authenticator.utils.AuthProperties;
import es.guadaltel.framework.authenticator.utils.UtilsWSPSIS;
import es.guadaltel.framework.core.Document;
import es.guadaltel.framework.core.User;
import es.guadaltel.framework.signer.SignTransaction;
import es.guadaltel.framework.signer.Signer;
import es.guadaltel.framework.signer.SignerFactory;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import javax.servlet.http.HttpServletRequest;
import net.sf.json.JSONArray;
import net.sf.json.JSONObject;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.InputStreamRequestEntity;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.log4j.Logger;
import sun.misc.BASE64Decoder;

public class PsisAuthenticatorImpl extends AfirmaAuthenticatorImpl
implements Authenticator, Serializable
{
	private static final long serialVersionUID = 603875688494009190L;
	private static Logger LOG = Logger.getLogger(PsisAuthenticatorImpl.class);

	public PsisAuthenticatorImpl(Configuration config)
	{
		super(config);
	}

	public User authenticate(HttpServletRequest request) {
		LOG.debug("Entrando en el metodo validarCertificadoPSIS");

		String certificate = null;

		User userDetails = new User();

		PostMethod method = null;

		LOG.debug("Recuperando certificado digital...");

		request.getSession().removeAttribute("clientSignerConfig");
		SignTransaction signTransaction = (SignTransaction)request.getSession().getAttribute("signTransAutenticacion");
		try
		{
			if (signTransaction == null) {
				LOG.error("Ocurrio un error al iniciar la autenticacion");
				throw new AuthenticatorException("Ocurrio un error al iniciar la autenticacion");
			}

			request.getSession().removeAttribute("signTransAutenticacion");

			String csResult = request.getParameter("clientSignerResult");

			if ((csResult == null) || (csResult.equals(""))) {
				LOG.error("Ocurrio un error en la obtencion de inputData.clientSignerResult");
				throw new AuthenticatorException("Ocurrio un error en la obtencion de inputData.clientSignerResult");
			}

			JSONArray jsonArray = JSONArray.fromObject(csResult);

			if ((jsonArray == null) || (jsonArray.isEmpty())) {
				LOG.error("Ocurrio un error en la obtencion de la informacion de firma.");
				throw new AuthenticatorException("Ocurrio un error en la obtencion de la informacion de firma.");
			}

			certificate = (String)((JSONObject)jsonArray.get(1)).get("certificate");

			if ((certificate == null) || (certificate.equals(""))) {
				LOG.error("Ocurrio un error en la obtencion del certificado.");
				throw new AuthenticatorException("Ocurrio un error en la obtencion del certificado.");
			}

			signTransaction.setUser(certificate);
			userDetails.setProperty("certificado", certificate);

			InfoCertificado infoCertificado = null;

			String url = AuthProperties.getProperty("authenticator.psis.endpoint");

			HttpClient client = new HttpClient();

			StringBuffer data = new StringBuffer();
			data.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
			data.append("<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\"><soapenv:Body>");
			data.append("<urn:VerifyRequest Profile=\"urn:oasis:names:tc:dss:1.0:profiles:XSS\" xmlns:urn=\"urn:oasis:names:tc:dss:1.0:core:schema\"><urn:OptionalInputs><urn:ReturnProcessingDetails/></urn:OptionalInputs><urn:SignatureObject><urn:Other><xd:X509Data xmlns:xd=\"http://www.w3.org/2000/09/xmldsig#\"><xd:X509Certificate>");
			data.append(certificate);
			data.append("</xd:X509Certificate></xd:X509Data></urn:Other></urn:SignatureObject></urn:VerifyRequest>");
			data.append("</soapenv:Body></soapenv:Envelope>");

			LOG.debug("[PSIS Request]...\n" + data.toString());

			ByteArrayInputStream in = new ByteArrayInputStream(data.toString().getBytes());

			method = new PostMethod(url);
			InputStreamRequestEntity isRequest = new InputStreamRequestEntity(in);
			method.setRequestEntity(isRequest);

			LOG.debug("Lanzando peticion...");
			int status = client.executeMethod(method);

			LOG.debug("Peticion finalizada, codigo de estado devuelto: " + status + " (OK=200)");
			if (status == 200)
			{
				InputStream stream = null;
				stream = method.getResponseBodyAsStream();
				ByteArrayOutputStream out = UtilsWSPSIS.getData(stream);

				String result = out.toString();
				LOG.debug("[PSIS Response]...\n" + result);

				if ((result.indexOf(AuthProperties.getProperty("authenticator.psis.majorResult")) > 0) && (result.indexOf(AuthProperties.getProperty("authenticator.psis.minorResult")) > 0)) {
					infoCertificado = getInfoCertificado(certificate);

					userDetails = UtilsWSPSIS.parseUser(infoCertificado);
				}
				else {
					LOG.error("Certificado no valido (caducado o revocado)");
				}
			} else {
				LOG.error("El servidor PSIS devolvio un estado incorrecto");
			}
		}
		catch (AuthenticatorException e) {
			e.printStackTrace();
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		finally
		{
			if (method != null) method.releaseConnection();
		}

		LOG.debug("Saliendo del metodo validarCertificadoPSIS");
		return userDetails;
	}

	private InfoCertificado getInfoCertificado(X509Certificate certificate)
	{
		InfoCertificado result = null;

		ICertReader reader = new CatCertReader();
		if (reader.isTypeOf(certificate)) {
			LOG.debug("Recuperando informacion de certificado CatCert");
			result = reader.getInfo(certificate);
		}
		else {
			reader = new FnmtCertReader();
			if (reader.isTypeOf(certificate)) {
				LOG.debug("Recuperando informacion de certificado FNMT");
				result = reader.getInfo(certificate);
			}
			else {
				reader = new DnieCertReader();
				if (reader.isTypeOf(certificate)) {
					LOG.debug("Recuperando informacion de certificado Dnie");
					result = reader.getInfo(certificate);
				}
				else {
					LOG.debug("Tipo de certificado no soportado");
				}
			}
		}
		if (result != null) {
			LOG.debug("Extracted subject: " + result.getSubject());
		}
		return result;
	}

	private InfoCertificado getInfoCertificado(String certificate)
	{
		X509Certificate cert = null;
		try
		{
			BASE64Decoder decoder = new BASE64Decoder();

			ByteArrayInputStream in = new ByteArrayInputStream(decoder.decodeBuffer(certificate));
			CertificateFactory factory = CertificateFactory.getInstance("X.509");
			cert = (X509Certificate)factory.generateCertificate(in);
		} catch (CertificateException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		return getInfoCertificado(cert);
	}

	public String getAuthenticationURL(HttpServletRequest request) throws AuthenticatorException {
		LOG.debug("Entrando en el metodo getAuthenticationURL");

		String retorno = "";
		try
		{
			String idSesion = request.getSession().getId();

			if ((idSesion == null) || (idSesion.equals(""))) {
				LOG.error("Ocurrio un error al obtener el id de sesion.");
				throw new AuthenticatorException("Ocurrio un error al obtener el id de sesion.");
			}

			Signer signer = SignerFactory.createInstance();
			SignTransaction signTransaction = new SignTransaction();

			Document document = new Document();
			document.setContent(idSesion.getBytes());

			LOG.debug("Realizamos firma en dos fases para guardar signTransaction en sesion");
			signTransaction.addDocument(document);
			signTransaction.setProperty("type", "twoPhases");
			signer.initSign(signTransaction, request);

			String clientSignerConfig = (String)request.getAttribute("clientSignerConfig");

			if ((clientSignerConfig == null) || (clientSignerConfig.equals(""))) {
				LOG.error("Ocurrio un error al obtener la configuracion del cliente de firma");
				throw new AuthenticatorException("Ocurrio un error al obtener la configuracion del cliente de firma");
			}

			request.getSession().setAttribute("clientSignerConfig", clientSignerConfig);
			request.getSession().setAttribute("signTransAutenticacion", signTransaction);

			retorno = AuthProperties.getProperty("authenticator.psis.url.return");

			if ((retorno == null) || (retorno.equals("")))
				LOG.warn("Ocurrio un error al obtener la url de retorno de autenticacion");
		}
		catch (Exception e)
		{
			LOG.error("Error in method getAuthenticationURL. Error: " + e.getMessage());
			throw new AuthenticatorException(e);
		}

		LOG.debug("Saliendo del metodo getAuthenticationURL");
		return retorno;
	}
}